package Linux::UVC;
use strict;
use warnings;

our $uvc_cmd='/usr/bin/uvcdynctrl';

sub devices
{
  my $class=shift // __PACKAGE__;
  my @lst;

  for my $dev (__read_devices())
  { push(@lst,$class->_new($dev)); }

  return @lst;
}

sub _new
{
  my $class=shift;
  my $data=shift;
  return undef unless -c $data->{device};

  my $ctrls=__read_controlls($data->{device});
  return undef unless $ctrls;
  for my $name (keys(%$ctrls))
  { $ctrls->{$name}=Linux::UVC::Controll->_new($data->{device},$name,$ctrls->{$name}); }

  $class=ref($class)? ref($class):$class;
  return bless({%$data, controlls => $ctrls },$class);
}

sub device { $_[0]->{device}; }
sub name { $_[0]->{name}; }
sub driver { $_[0]->{driver}; }
sub id { $_[0]->{id}; }

sub controlls{ return sort{ $a->name() cmp $b->name() }values(%{$_[0]->{controlls}}) }
sub controll_names{ return keys(%{$_[0]->{controlls}}) }
sub controll_by_name
{
  my $self=shift;
  my $name=shift // '';
  return undef unless(exists($self->{controlls}->{$name}));
  return $self->{controlls}->{$name};
}

#-----------------------------------------------------------------------

sub __read_devices
{
  open(my $ph, '-|', $Linux::UVC::uvc_cmd.' -v -l 2> /dev/null') or return undef;
  my @ret;
  while(my $line=<$ph>)
  {
    chomp($line);
    if($line=~/\s+(.+?)\s+(.+?)\s+\[(.+?),\s+(.+?)\]$/)
    { push(@ret,{ device=>"/dev/$1", name => $2, driver => $3, id=>$4 }); }
  }
  close($ph);
  return @ret;
}

sub __read_controlls
{
  my $device=shift;
  open(my $ph, '-|', $Linux::UVC::uvc_cmd." -d '$device' -v -c 2> /dev/null") or return undef;
  my $ret={};
  my $now={};
  while(my $line=<$ph>)
  {
    chomp($line);
    if($line=~/^\s+(\w+?)$/)
    {
      $now={};
      $ret->{$1}=$now;
    }
    elsif($line=~/^\s+(.+?)\s+:\s+(.+?),?$/)
    {
      my $name=$1;
      my $val=$2;
      if($name eq 'Values')
      {
        if($val=~/\[ (\d+) \.\. (\d+), step size: (\d+) \]/)
        { $val={min => $1, max => $2, step => $3}; }
      }
      elsif($name eq 'Flags')
      {
        if($val=~/\{ (.+?) \}/)
        {
          $val={};
          for my $v (split(/\s*,\s*/, $1))
          { $val->{$v}=1; }
        }
      }

      $now->{$name}=$val;
    }
  }
  close($ph) or return undef;
  return $ret;
}
1;

########################################################################
package Linux::UVC::Controll;

sub _new
{
  my $class=shift;
  my $device=shift;
  my $name=shift;
  my $vals=shift;
  $class=ref($class)? ref($class):$class;
  bless({device => $device, values => $vals, name => $name},$class);
}

sub max{ $_[0]->{values}->{Values}->{max}; }
sub min{ $_[0]->{values}->{Values}->{min}; }
sub step{ $_[0]->{values}->{Values}->{step}; }
sub default{ $_[0]->{values}->{Default}; }
sub name{ $_[0]->{name}; }
sub id{ $_[0]->{values}->{ID}; }
sub type{ $_[0]->{values}->{Type}; }
sub flags{ $_[0]->{values}->{Flags}; }

sub set
{
  my $self=shift;
  my $val=shift;
  $val=$self->max() if($val > $self->max());
  $val=$self->min() if($val < $self->min());
  return 0 unless($self->flags()->{CAN_WRITE});
  return __write_controll($self->{device},$self->{name},$val);
}

sub get
{
  my $self=shift;
  return undef unless($self->flags()->{CAN_READ});
  return __read_controll($self->{device},$self->{name});
}

sub value
{
  my $self=shift;
  if(@_)
  { $self->set(@_); }
  return $self->get();
}

sub inc
{
  my $self=shift;
  my $val=$self->get();
  return undef unless(defined($val));
  $val+=$self->step();
  return undef unless($self->set($val));
  return $val;
}

sub dec
{
  my $self=shift;
  my $val=$self->get();
  return undef unless(defined($val));
  $val-=$self->step();
  return undef unless($self->set($val));
  return $val;
}

sub reset
{
  my $self=shift;
  $self->set($self->default());
}

#-----------------------------------------------------------------------

sub __read_controll
{
  my $device=shift;
  my $controll=shift;

  my $val=undef;

  open(my $ph, '-|', $Linux::UVC::uvc_cmd." -d '$device' -g '$controll' 2> /dev/null") or return undef;
  chomp($val=<$ph>);
  close($ph) or return undef;

  return $val;
}

sub __write_controll
{
  my $device=shift;
  my $controll=shift;
  my $val=shift;

  return system($Linux::UVC::uvc_cmd." -d '$device' -s '$controll' -- '$val' 2> /dev/null") == 0;
}

1;